import torch 

'''
adapted from Jin Chen's hotrg.jl at https://github.com/wangleiphy/TRG/blob/master/HOTRG.jl 
'''

def gauge(T, Dcut, side):
    #t0=time.time()
    D = T.shape[0]
    M = torch.einsum('ludr,adcb->laucrb', (T, T)).contiguous()
    if side=='l':
        M = M.view(D**2, -1)
        M = M@M.t()
    elif side=='r':
        M = M.view(-1, D**2)
        M = M.t()@M
    #print ('build M done {:.3f}s'.format(time.time()-t0))

    #t0=time.time()
    U, S, V = torch.svd(M)
    #print ('svd M done {:.3f}s'.format(time.time()-t0))
    Dnew = min(D**2, Dcut)
    truncation = S[Dnew:].sum() / S.sum()
    return U[:, :Dnew].view(D, D, Dnew), truncation

def hotrg(T, D, Dcut, no_iter):
    lnZ = 0.0
    truncation_error = 0.0
    for n in range(no_iter):
        f = torch.norm(T)
        lnZ += 2**(no_iter-n)*torch.log(f)
        T = T / f

        #contract up-down
        vl, el = gauge(T, Dcut,'l')
        vr, er = gauge(T, Dcut,'r')

        U = vl if (el < er) else vr
        truncation_error += min(el, er)
        #t0=time.time()
        T = torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))
        #print ('truncate done {:.3f}s'.format(time.time()-t0))
       
    trace = 0.0 
    for x in range(T.shape[0]):
        for y in range(T.shape[1]):
            trace += T[x, y, y, x]
    lnZ += torch.log(trace)
    return lnZ/2**no_iter, truncation_error

if __name__=='__main__':
    import time
    K = torch.tensor([0.44])
    Dcut = 20
    n = 20

    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

    lnZ, _ = hotrg(T, 2, Dcut, n)
    print (lnZ.item())
